class TreeNodeBuilder
  include MiqAeClassHelper

  # method to build non-explorer tree nodes
  def self.generic_tree_node(key, text, image, tip = nil, options = {})
    text = ERB::Util.html_escape(text) unless text.html_safe?
    node = {
      :key   => key,
      :title => text,
    }
    node[:icon]         = ActionController::Base.helpers.image_path("100/#{image}") if image
    node[:addClass]     = options[:style_class]      if options[:style_class]
    node[:cfmeNoClick]  = true                       if options[:cfme_no_click]
    node[:expand]       = options[:expand]           if options[:expand]
    node[:hideCheckbox] = true                       if options[:hideCheckbox]
    node[:noLink]       = true                       if options[:noLink]
    node[:select]       = options[:select]           if options[:select]
    node[:tooltip]      = ERB::Util.html_escape(tip) if tip && !tip.html_safe?
    node
  end

  # Options used:
  #   :type       -- Type of tree, i.e. :handc, :vandt, :filtered, etc
  #   :open_nodes -- Tree node ids of currently open nodes
  #   FIXME: fill in missing docs
  #
  def self.build(object, parent_id, options)
    builder = new(object, parent_id, options)
    builder.build
  end

  def self.build_id(object, parent_id, options)
    builder = new(object, parent_id, options)
    builder.build_id
  end

  def initialize(object, parent_id, options)
    @object, @parent_id, @options = object, parent_id, options
  end

  attr_reader :object, :parent_id, :options

  def build_id
    object.kind_of?(Hash) ? build_hash_id : build_object_id
  end

  def build
    case object
    when AssignedServerRole   then assigned_server_role_node(object)
    when AvailabilityZone     then generic_node(object.name,
                                                "availability_zone.png",
                                                _("Availability Zone: %{name}") % {:name => object.name})
    when ConfigurationScript then generic_node(object.name,
                                               "configuration_script.png",
                                               "Ansible Tower Job Template: #{object.name}")
    when ExtManagementSystem  then
      # TODO: This should really leverage .base_model on an EMS
      prefix_model =
        case object
        when EmsCloud then "EmsCloud"
        when EmsInfra then "EmsInfra"
        else               "ExtManagementSystem"
        end

      generic_node(object.name, "vendor-#{object.image_name}.png", "#{ui_lookup(:model => prefix_model)}: #{object.name}")
    when ChargebackRate       then generic_node(object.description, "chargeback_rate.png")
    when Classification       then classification_node
    when Compliance
      name = "<b>" + _("Compliance Check on: ") + "</b>" + format_timezone(object.timestamp, Time.zone, 'gtl')
      generic_node(name.html_safe, "#{object.compliant ? "check" : "x"}.png")
    when ComplianceDetail
      name = "<b>" + _("Policy: ") + "</b>" + object.miq_policy_desc
      generic_node(name.html_safe, "#{object.miq_policy_result ? "check" : "x"}.png")
    when Condition            then generic_node(object.description, "miq_condition.png")
    when ConfigurationProfile then configuration_profile_node(object.name, "configuration_profile.png",
                                                              _("Configuration Profile: %{name}") % {:name => object.name})
    when ConfiguredSystem     then generic_node(object.hostname,
                                                "configured_system.png",
                                                _("Configured System: %{hostname}") % {:hostname => object.hostname})
    when Container            then generic_node(object.name, "container.png")
    when CustomButton         then generic_node(object.name,
                                                object.options && object.options[:button_image] ? "custom-#{object.options[:button_image]}.png" : "leaf.gif",
                                                _("Button: %{button_description}") % {:button_description => object.description})
    when CustomButtonSet      then custom_button_set_node
    when CustomizationTemplate then generic_node(object.name, "customizationtemplate.png")
    when Dialog               then generic_node(object.label, "dialog.png")
    when DialogTab            then generic_node(object.label, "dialog_tab.png")
    when DialogGroup          then generic_node(object.label, "dialog_group.png")
    when DialogField          then generic_node(object.label, "dialog_field.png")
    when EmsFolder            then ems_folder_node
    when EmsCluster           then generic_node(object.name, "cluster.png", "#{ui_lookup(:table => "ems_cluster")}: #{object.name}")
    when GuestDevice          then guest_node(object)
    when Host                 then generic_node(object.name,
                                                "host.png",
                                                "#{ui_lookup(:table => "host")}: #{object.name}")
    when IsoDatastore         then generic_node(object.name, "isodatastore.png")
    when IsoImage             then generic_node(object.name, "isoimage.png")
    when ResourcePool         then generic_node(object.name, object.vapp ? "vapp.png" : "resource_pool.png")

    when Vm                   then vm_node(object)
    when Lan                  then generic_node(object.name,
                                                "lan.png",
                                                _("Port Group: %{name}") % {:name => object.name})
    when LdapDomain           then generic_node(_("Domain: %{domain_name}") % {:domain_name => object.name},
                                                "ldap_domain.png",
                                                _("LDAP Domain: %{ldap_domain_name}") % {:ldap_domain_name => object.name})
    when LdapRegion           then generic_node(_("Region: %{region_name}") % {:region_name => object.name},
                                                "ldap_region.png",
                                                _("LDAP Region: %{ldap_region_name}") % {:ldap_region_name => object.name})
    when MiqAeClass           then node_with_display_name("ae_class.png")
    when MiqAeInstance        then node_with_display_name("ae_instance.png")
    when MiqAeMethod          then node_with_display_name("ae_method.png")
    when MiqAeNamespace       then node_with_display_name("ae_namespace.png")
    when MiqAlertSet          then generic_node(object.description, "miq_alert_profile.png")
    when MiqReport            then generic_node(object.name, "report.png")
    when MiqReportResult      then miq_report_node(object.last_run_on, object.name, object.status)
    when MiqSchedule          then generic_node(object.name, "miq_schedule.png")
    when MiqScsiLun           then generic_node(object.canonical_name,
                                                "lun.png",
                                                _("LUN: %{name}") % {:name => object.canonical_name})
    when MiqScsiTarget        then miq_scsi_target(object.iscsi_name, object.target)
    when MiqServer            then miq_server_node
    when MiqTemplate          then generic_node(object.name, "currentstate-#{object.normalized_state.downcase}.png")
    when MiqAlert             then generic_node(object.description, "miq_alert.png")
    when MiqAction            then miq_action_node
    when MiqEventDefinition   then generic_node(object.description, "event-#{object.name}.png")
    when MiqGroup             then generic_node(object.name, "group.png")
    # Following line has dynatree workaround, add class to allow clicking on bold portion of title.
    when MiqPolicy            then miq_policy_node
    when MiqPolicySet         then generic_node(object.description, "policy_profile#{object.active? ? "" : "_inactive"}.png")
    when MiqUserRole          then generic_node(object.name, "miq_user_role.png")
    when OrchestrationTemplateCfn then generic_node(object.name, "orchestration_template_cfn.png")
    when OrchestrationTemplateHot then generic_node(object.name, "orchestration_template_hot.png")
    when OrchestrationTemplateAzure then generic_node(object.name, "orchestration_template_azure.png")
    when OrchestrationTemplateVnfd then generic_node(object.name, "orchestration_template_vnfd.png")
    when ManageIQ::Providers::Vmware::CloudManager::OrchestrationTemplate then generic_node(object.name, "orchestration_template_vapp.png")
    when PxeImage             then generic_node(object.name, object.default_for_windows ? "win32service.png" : "pxeimage.png")
    when WindowsImage         then generic_node(object.name, "os-windows_generic.png")
    when PxeImageType         then generic_node(object.name, "pxeimagetype.png")
    when PxeServer            then generic_node(object.name, "pxeserver.png")
    when ScanItemSet          then generic_node(object.name, "scan_item_set.png")
    when Service              then generic_node(object.name, object.picture ? "/pictures/#{object.picture.basename}" : "service.png")
    when ServiceResource      then generic_node(object.resource_name, object.resource_type == "VmOrTemplate" ? "vm.png" : "service_template.png")
    when ServerRole           then server_role_node(object)
    when ServiceTemplate      then service_template_node
    when ServiceTemplateCatalog then service_template_catalog_node
    when Snapshot             then snapshot_node
    when Storage              then generic_node(object.name, "storage.png")
    when Switch               then generic_node(object.name,
                                                "switch.png",
                                                _("Switch: %{name}") % {:name => object.name})
    when User                 then generic_node(object.name, "user.png")
    when MiqSearch            then generic_node(object.description,
                                                "filter.png",
                                                _("Filter: %{filter_description}") % {:filter_description => object.description})
    when MiqDialog            then generic_node(object.description, "miqdialog.png", object[0])
    when MiqRegion            then miq_region_node
    when MiqWidget            then generic_node(object.title, "#{object.content_type}_widget.png", object.title)
    when MiqWidgetSet         then generic_node(object.name, "dashboard.png", object.name)
    when Tenant               then generic_node(object.name,  "#{object.tenant? ? "tenant" : "project"}.png")
    when VmdbTableEvm         then generic_node(object.name, "vmdbtableevm.png")
    when VmdbIndex            then generic_node(object.name, "vmdbindex.png")
    when Zone                 then zone_node
    when Hash                 then hash_node
    end
    @node
  end

  private

  def get_rr_status_image(status)
    case status
    when 'error'    then 'report_result_error.png'
    when 'finished' then 'report_result_completed.png'
    when 'running'  then 'report_result_running.png'
    when 'queued'   then 'report_result_queued.png'
    else                 'report_result.png'
    end
  end

  def tooltip(tip)
    unless tip.blank?
      tip = tip.kind_of?(Proc) ? tip.call : _(tip)
      tip = ERB::Util.html_escape(URI.unescape(tip)) unless tip.html_safe?
      @node[:tooltip] = tip
    end
  end

  def node_icon(icon)
    if icon.start_with?("/")
      icon
    else
      ActionController::Base.helpers.image_path("100/#{icon}")
    end
  end

  def generic_node(text, image, tip = nil)
    text = ERB::Util.html_escape(text ? URI.unescape(text) : text) unless text.html_safe?
    @node = {
      :key   => build_object_id,
      :title => text,
      :icon  => node_icon(image)
    }
    # Start with all nodes open unless expand is explicitly set to false
    @node[:expand] = options[:open_all].present? && options[:open_all] && options[:expand] != false
    @node[:hideCheckbox] = options[:hideCheckbox] if options[:hideCheckbox].present?
    tooltip(tip)
  end

  def normal_folder_node
    icon = options[:type] == :vandt ? "blue_folder.png" : "folder.png"
    generic_node(object.name, icon, _("Folder: %{folder_name}") % {:folder_name => object.name})
  end

  def guest_node(object)
    if object.device_type == "ethernet"
      generic_node(object.device_name, "pnic.png", _("Physical NIC: %{name}") % {:name => object.device_name})
    else
      generic_node(object.device_name,
                   "sa_#{object.controller_type.downcase}.png",
                   _("%{type} Storage Adapter: %{name}") % {:type => object.controller_type,
                                                            :name => object.device_name})
    end
  end

  def classification_node
    generic_node(object.description, 'folder', _("Category: %{description}") % {:description => object.description})
    @node[:cfmeNoClick] = true
    @node[:hideCheckbox] = true
  end

  def hash_node
    text = object[:text]
    text = text.kind_of?(Proc) ? text.call : _(text)

    # FIXME: expansion
    @node = {
      :key   => build_hash_id,
      :title => ERB::Util.html_escape(text)
    }
    @node[:icon] = node_icon("#{object[:image] || text}.png") if object[:image]
    # Start with all nodes open unless expand is explicitly set to false
    @node[:expand] = true if options[:open_all] && options[:expand] != false
    @node[:cfmeNoClick] = object[:cfmeNoClick] if object.key?(:cfmeNoClick)
    @node[:hideCheckbox] = true if object.key?(:hideCheckbox)
    @node[:select] = object[:select] if object.key?(:select)
    @node[:addClass] = object[:addClass] if object.key?(:addClass)
    @node[:checkable] = object[:checkable] if object.key?(:checkable)

    # FIXME: check the following
    # TODO: With dynatree, unless folders are open, we can't jump to a child node until it has been visible once
    # node[:expand] = false

    tooltip(object[:tip])
  end

  def node_with_display_name(image)
    text = object.display_name.blank? ? object.name : "#{object.display_name} (#{object.name})"
    if object.kind_of?(MiqAeNamespace) && object.domain?
      editable_domain = editable_domain?(object)
      enabled_domain  = object.enabled
      unless editable_domain && enabled_domain
        text = add_read_only_suffix(text, editable_domain, enabled_domain)
        return miq_ae_node(enabled_domain,
                           text,
                           image_for_node(object, image),
                           "#{tooltip_prefix_for_node(object)}: #{text}"
                          )
      end
    end
    generic_node(text, image_for_node(object, image), "#{tooltip_prefix_for_node(object)}: #{text}")
  end

  def miq_ae_node(enabled, text, image, tip)
    text = ERB::Util.html_escape(text) unless text.html_safe?
    @node = {
      :key   => build_object_id,
      :title => text,
      :icon  => node_icon(image)
    }
    @node[:addClass] = "striketrough" unless enabled
    @node[:expand] = true if options[:open_all]  # Start with all nodes open
    tooltip(tip)
  end

  def configuration_profile_node(text, image, tip = nil)
    text = ERB::Util.html_escape(text) unless text.html_safe?
    title = text.split('|').first
    @node = {
      :key   => build_object_id,
      :title => title,
      :icon  => node_icon(title == _("Unassigned Profiles Group") ? "folder.png" : image)
    }
    @node[:expand] = true if options[:open_all]  # Start with all nodes open
    tooltip(tip)
  end

  def vm_node(object)
    image = "currentstate-#{object.normalized_state.downcase}.png"
    if object.template?
      image = object.host ? "template.png" : "template-no-host.png"
    end
    generic_node(object.name, image, _("VM: %{name} (Click to view)") % {:name => object.name})
  end

  def image_for_node(object, image)
    case object
    when MiqAeNamespace
      if object.domain?
        object.git_enabled? ? "ae_git_domain.png" : domain_png(object)
      else
        "ae_namespace.png"
      end
    else
      image
    end
  end

  def domain_png(object)
    return 'miq.png' if object.name == MiqAeDatastore::MANAGEIQ_DOMAIN
    object.top_level_namespace ? "vendor-#{object.top_level_namespace.downcase}.png" : "ae_domain.png"
  end

  def tooltip_prefix_for_node(object)
    case object
    when MiqAeNamespace
      object.domain? ? ui_lookup(:model => "MiqAeDomain") : ui_lookup(:model => object.class.to_s)
    else
      ui_lookup(:model => object.class.to_s)
    end
  end

  def custom_button_set_node
    text = if options[:type] == :sandt
             _("%{button_group_name} (Group)") % {:button_group_name => object.name.split("|").first}
           else
             object.name.split("|").first
           end
    image = object.set_data && object.set_data[:button_image] ? "custom-#{object.set_data[:button_image]}.png" : "folder.png"
    tip = if object.description
            _("Button Group: %{button_group_description}") % {:button_group_description => object.description}
          else
            object.name.split("|").first
          end
    generic_node(text, image, tip)
  end

  def ems_folder_node
    if object.kind_of?(Datacenter)
      generic_node(object.name,
                   "datacenter.png",
                   _("Datacenter: %{datacenter_name}") % {:datacenter_name => object.name})
    else # normal Folders
      normal_folder_node
    end
  end

  def miq_scsi_target(iscsi_name, target)
    name = if iscsi_name.blank?
             _("SCSI Target %{target}") % {:target => target}
           else
             _("SCSI Target %{target} (%{name})") % {:target => target, :name => iscsi_name}
           end
    generic_node(name, "target_scsi.png", _("Target: %{text}") % {:text => name})
  end

  def miq_server_node
    if options[:is_current]
      tip  = _("%{server}: %{server_name} [%{server_id}] (current)") %
             {:server => ui_lookup(:model => object.class.to_s), :server_name => object.name, :server_id => object.id}
      tip += " (#{object.status})" if options[:tree] == :roles_by_server_tree
      text = "<strong>#{ERB::Util.html_escape(tip)}</strong>".html_safe
    else
      tip  = "#{ui_lookup(:model => object.class.to_s)}: #{object.name} [#{object.id}]"
      tip += " (#{object.status})" if options[:tree] == :roles_by_server_tree
      text = tip
    end
    generic_node(text, 'miq_server.png', tip)
    @node[:expand] = true
    @node
  end

  def miq_action_node
    if options[:tree] != :action_tree
      if options[:tree] == :policy_profile_tree
        policy_id = parent_id.split('-')[2].split('_').first
        event_id  = parent_id.split('-').last
      else
        policy_id = parent_id.split('_')[2].split('-').last
        event_id  = parent_id.split('_').last.split('-').last
      end
      p  = MiqPolicy.find_by_id(ApplicationRecord.uncompress_id(policy_id))
      ev = MiqEventDefinition.find_by_id(ApplicationRecord.uncompress_id(event_id))
      image = p.action_result_for_event(object, ev) ? "check" : "x"
    else
      image = object.action_type == "default" ? "miq_action" : "miq_action_#{object.action_type}"
    end
    generic_node(object.description, "#{image}.png")
  end

  def miq_region_node
    generic_node(object.name, "miq_region.png", object[0])
    @node[:expand] = true
    @node
  end

  def service_template_node
    generic_node(object.name, object.picture ? "../../../pictures/#{object.picture.basename}" : "service_template.png")
    @node[:title] += " (%s)" % object.tenant.name unless object.tenant.ancestors.empty?
  end

  def service_template_catalog_node
    generic_node(object.name, "service_template_catalog.png")
    @node[:title] += " (%s)" % object.tenant.name if object.tenant.present? && object.tenant.ancestors.present?
  end

  def snapshot_node
    generic_node(object.name, 'snapshot.png', object.name)
    @node[:title] += _(' (Active)') if object.current?
  end

  def zone_node
    if options[:is_current]
      tip  = _("%{zone}: %{zone_description} (current)") %
             {:zone => ui_lookup(:model => object.class.to_s), :zone_description => object.description}
      text = "<strong>#{ERB::Util.html_escape(tip)}</strong>".html_safe
    else
      tip  = "#{ui_lookup(:model => object.class.to_s)}: #{object.description}"
      text = tip
    end
    generic_node(text, "zone.png", tip)
  end

  def policy_profile_text
    ["<strong>", ui_lookup(:model => object.towhat),
     " ", object.mode.titleize, ":</strong> ",
     ERB::Util.html_escape(object.description)].join('').html_safe
  end

  def miq_policy_node
    text  = options[:tree] == :policy_profile_tree ? policy_profile_text : object.description
    image = "miq_policy_#{object.towhat.downcase}#{object.active ? '' : '_inactive'}.png"
    generic_node(text, image)
  end

  def format_parent_id
    (options[:full_ids] && !parent_id.blank?) ? "#{parent_id}_" : ''
  end

  def build_hash_id
    if object[:id] == "-Unassigned"
      "-Unassigned"
    else
      prefix = TreeBuilder.get_prefix_for_model("Hash")
      "#{format_parent_id}#{prefix}-#{object[:id]}"
    end
  end

  def build_object_id
    if object.id.nil?
      # FIXME: this makes problems in tests
      # to handle "Unassigned groups" node in automate buttons tree
      "-#{object.name.split('|').last}"
    else
      base_class = object.class.base_model.name           # i.e. Vm or MiqTemplate
      base_class = "Datacenter" if base_class == "EmsFolder" && object.kind_of?(Datacenter)
      base_class = "ManageIQ::Providers::Foreman::ConfigurationManager" if object.kind_of?(ManageIQ::Providers::Foreman::ConfigurationManager)
      base_class = "ManageIQ::Providers::AnsibleTower::ConfigurationManager" if object.kind_of?(ManageIQ::Providers::AnsibleTower::ConfigurationManager)
      prefix = TreeBuilder.get_prefix_for_model(base_class)
      cid = ApplicationRecord.compress_id(object.id)
      "#{format_parent_id}#{prefix}-#{cid}"
    end
  end

  def miq_report_node(last_run_on, name, status)
    status = status.downcase
    image = get_rr_status_image(status)
    if last_run_on.nil? && (status == "queued" || status == "running")
      expand = !!options[:open_all]

      @node = TreeNodeBuilder.generic_tree_node(
        build_object_id,
        _("Generating Report"),
        image,
        _("Generating Report for - %{report_name}") % {:report_name => name},
        :expand => expand
      )
    elsif last_run_on.nil? && status == "error"
      generic_node(_("Error Generating Report"), image,
        _("Error Generating Report for %{report_name}") % {:report_name => name})
    else
      text = format_timezone(last_run_on, Time.zone, 'gtl')
      generic_node(text, image)
    end
  end

  def assigned_server_role_node(object)
    @node = {
      :key   => build_object_id,
      :title => options[:tree] == :servers_by_role_tree ?
        "<strong>#{_('Server')}: #{object.name} [#{object.id}]</strong>" :
        "<strong>Role: #{object.server_role.description}</strong>"
    }

    if object.master_supported?
      priority = case object.priority
                 when 1
                   _("primary, ")
                 when 2
                   _("secondary, ")
                 else
                   ""
                 end
    end
    if object.active? && object.miq_server.started?
      @node[:icon] = ActionController::Base.helpers.image_path("100/on.png")
      @node[:title] += _(" (%{priority}active, PID=%{number})") % {:priority => priority, :number => object.miq_server.pid}
    else
      if object.miq_server.started?
        @node[:icon] = ActionController::Base.helpers.image_path("100/suspended.png")
        @node[:title] += _(" (%{priority}available, PID=%{number})") % {:priority => priority,
                                                                        :number   => object.miq_server.pid}
      else
        @node[:icon] = ActionController::Base.helpers.image_path("100/off.png")
        @node[:title] += _(" (%{priority}unavailable)") % {:priority => priority}
      end
      @node[:addClass] = "red" if object.priority == 1
    end
    if @options[:parent_kls] == "Zone" && object.server_role.regional_role?
      @node[:addClass] = "opacity"
    end
    @node
  end

  def server_role_node(object)
    status = "stopped"
    object.assigned_server_roles.where(:active => true).each do |asr| # Go thru all active assigned server roles
      next unless asr.miq_server.started? # Find a started server
      if @options[:parent_kls] == "MiqRegion" || # it's in the region
         (@options[:parent_kls] == "Zone" && asr.miq_server.my_zone == @options[:parent_name]) # it's in the zone
        status = "active"
        break
      end
    end
    @node = {
      :key    => build_object_id,
      :title  => _("Role: %{description} (%{status})") % {:description => object.description, :status => status},
      :icon   => ActionController::Base.helpers.image_path("100/role-#{object.name}.png"),
      :expand => true
    }
    tooltip(_("Role: %{description} (%{status})") % {:description => object.description, :status => status})
    @node
  end
end
